import { MinecraftBlockTypes, BlockPermutation } from "@minecraft/server";
import { Vector } from "../utils/vector";
import { Server } from "./serverBuilder";
export class BlockBuilder {
    /**
     * Converts block data value to block states.
     * @param block the block identifier
     * @param data the data value
     * @returns an object with block state properties and corresponding values
     */
    dataValueToStates(block, data) {
        if (!block.startsWith("minecraft:")) {
            block = "minecraft:" + block;
        }
        if (block in dataToStates) {
            const states = dataToStates[block];
            return states[data >= states.length ? 0 : data];
        }
        return {};
    }
    dataValueToPermutation(block, data) {
        if (!block.startsWith("minecraft:")) {
            block = "minecraft:" + block;
        }
        return BlockPermutation.resolve(block, this.dataValueToStates(block, data));
    }
    isAirOrFluid(block) {
        if (!block)
            return true;
        const type = block.type.id;
        return type == "minecraft:air" || type == "minecraft:water" || type == "minecraft:lava";
    }
}
export const Block = new BlockBuilder();
/**
 * @deprecated
 * Call this function to get the maps between data value and block states (dataToStates)
 * Results are in console log, not warning, so it won't show in content log gui.
 * Use the script debugger to immediately copy the output from your IDE (eg: Vscode).
 */
// eslint-disable-next-line @typescript-eslint/no-unused-vars
async function fillDataMap(loc, dim) {
    let printed = false;
    const data = {};
    const blocks = MinecraftBlockTypes.getAllBlockTypes();
    const genGlass = (loc) => {
        const glass = BlockPermutation.resolve("minecraft:glass");
        for (const offset of [[-1, 0, 0], [1, 0, 0], [0, -1, 0], [0, 1, 0], [0, 0, -1], [0, 0, 1]]) {
            dim.getBlock(loc.add([offset[0], offset[1], offset[2]])).setPermutation(glass);
        }
    };
    const testBlocks = async (loc) => {
        genGlass(loc);
        let blockType;
        // eslint-disable-next-line no-cond-assign
        while (blockType = blocks.pop()) {
            const defaultPerm = BlockPermutation.resolve(blockType.id);
            if (!Object.keys(defaultPerm.getAllStates()).length) {
                continue;
            }
            const perms = [];
            let lastPerm = "";
            let dupeCount = 1;
            for (let i = 0; i < 64; i++) {
                const cmdResult = await Server.queueCommand(`setblock ${Vector.from(loc).print()} ${blockType.id} ${i}`);
                if (cmdResult.error) {
                    break;
                }
                const block = dim.getBlock(loc);
                if (block.isAir() && i == 0) {
                    break;
                }
                const props = block.permutation.getAllStates();
                for (const name in props) {
                    if (name == "persistent_bit") {
                        props[name] = true;
                    }
                }
                perms.push(props);
                genGlass(loc);
                const propString = JSON.stringify(props);
                if (lastPerm == propString) {
                    if (++dupeCount >= 4) {
                        break;
                    }
                }
                else {
                    lastPerm = propString;
                    dupeCount = 1;
                }
            }
            if (perms.length > 1) {
                const dupe = JSON.stringify(perms[perms.length - 1]);
                for (let i = perms.length - 2; i >= 0; i--) {
                    if (JSON.stringify(perms[i]) == dupe && i != 0)
                        continue;
                    if (i != perms.length - 2) {
                        perms.length = i + 1;
                    }
                    break;
                }
                if (perms.length > 1 && JSON.stringify(perms[0]) == JSON.stringify(perms[perms.length - 1])) {
                    perms.length--;
                }
                if (perms.length > 1) {
                    data[blockType.id] = JSON.stringify(perms);
                }
            }
        }
        if (!printed) {
            printed = true;
            console.log(JSON.stringify(data, null, "  "));
        }
    };
    for (let x = 0; x <= 8; x += 2)
        for (let z = 0; z <= 8; z += 2) {
            testBlocks(loc.add([x, 0, z]));
        }
}
const dataToStates = {
    "minecraft:bamboo_sapling": [{ "sapling_type": "oak", "age_bit": false }, { "sapling_type": "oak", "age_bit": true }],
    "minecraft:waxed_cut_copper_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:red_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:white_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:stripped_mangrove_log": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:cobbled_deepslate_wall": [{ "wall_connection_type_west": "tall", "wall_connection_type_south": "tall", "wall_connection_type_east": "tall", "wall_connection_type_north": "tall", "wall_post_bit": true }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false }],
    "minecraft:gray_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:polished_blackstone_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:cobbled_deepslate_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:lit_smoker": [{ "facing_direction": "south" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "facing_direction": "down" }],
    "minecraft:stonecutter_block": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:dispenser": [{ "triggered_bit": false, "facing_direction": "down" }, { "triggered_bit": false, "facing_direction": "up" }, { "triggered_bit": false, "facing_direction": "north" }, { "triggered_bit": false, "facing_direction": "south" }, { "triggered_bit": false, "facing_direction": "west" }, { "triggered_bit": false, "facing_direction": "east" }],
    "minecraft:blast_furnace": [{ "facing_direction": "south" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "facing_direction": "down" }],
    "minecraft:purpur_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:smooth_quartz_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:brewing_stand": [{ "brewing_stand_slot_c_bit": false, "brewing_stand_slot_b_bit": false, "brewing_stand_slot_a_bit": false }, { "brewing_stand_slot_c_bit": false, "brewing_stand_slot_b_bit": false, "brewing_stand_slot_a_bit": true }, { "brewing_stand_slot_c_bit": false, "brewing_stand_slot_b_bit": true, "brewing_stand_slot_a_bit": false }, { "brewing_stand_slot_c_bit": false, "brewing_stand_slot_b_bit": true, "brewing_stand_slot_a_bit": true }, { "brewing_stand_slot_c_bit": true, "brewing_stand_slot_b_bit": false, "brewing_stand_slot_a_bit": false }, { "brewing_stand_slot_c_bit": true, "brewing_stand_slot_b_bit": false, "brewing_stand_slot_a_bit": true }, { "brewing_stand_slot_c_bit": true, "brewing_stand_slot_b_bit": true, "brewing_stand_slot_a_bit": false }, { "brewing_stand_slot_c_bit": true, "brewing_stand_slot_b_bit": true, "brewing_stand_slot_a_bit": true }],
    "minecraft:chain": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:dropper": [{ "triggered_bit": false, "facing_direction": "down" }, { "triggered_bit": false, "facing_direction": "up" }, { "triggered_bit": false, "facing_direction": "north" }, { "triggered_bit": false, "facing_direction": "south" }, { "triggered_bit": false, "facing_direction": "west" }, { "triggered_bit": false, "facing_direction": "east" }],
    "minecraft:dark_prismarine_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:fence": [{ "wood_type": "oak" }, { "wood_type": "spruce" }, { "wood_type": "birch" }, { "wood_type": "jungle" }, { "wood_type": "acacia" }, { "wood_type": "dark_oak" }],
    "minecraft:lantern": [{ "hanging": false }, { "hanging": true }],
    "minecraft:warped_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:chest": [{ "facing_direction": "north" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "facing_direction": "down" }],
    "minecraft:purple_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:lit_blast_furnace": [{ "facing_direction": "south" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "facing_direction": "down" }],
    "minecraft:light_block": [{ "block_light_level": 0 }, { "block_light_level": 1 }, { "block_light_level": 2 }, { "block_light_level": 3 }, { "block_light_level": 4 }, { "block_light_level": 5 }, { "block_light_level": 6 }, { "block_light_level": 7 }, { "block_light_level": 8 }, { "block_light_level": 9 }, { "block_light_level": 10 }, { "block_light_level": 11 }, { "block_light_level": 12 }, { "block_light_level": 13 }, { "block_light_level": 14 }, { "block_light_level": 15 }],
    "minecraft:wooden_pressure_plate": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:turtle_egg": [{ "cracked_state": "no_cracks", "turtle_egg_count": "one_egg" }, { "cracked_state": "no_cracks", "turtle_egg_count": "two_egg" }, { "cracked_state": "no_cracks", "turtle_egg_count": "three_egg" }, { "cracked_state": "no_cracks", "turtle_egg_count": "four_egg" }, { "cracked_state": "cracked", "turtle_egg_count": "one_egg" }, { "cracked_state": "cracked", "turtle_egg_count": "two_egg" }, { "cracked_state": "cracked", "turtle_egg_count": "three_egg" }, { "cracked_state": "cracked", "turtle_egg_count": "four_egg" }, { "cracked_state": "max_cracked", "turtle_egg_count": "one_egg" }, { "cracked_state": "max_cracked", "turtle_egg_count": "two_egg" }, { "cracked_state": "max_cracked", "turtle_egg_count": "three_egg" }, { "cracked_state": "max_cracked", "turtle_egg_count": "four_egg" }, { "cracked_state": "no_cracks", "turtle_egg_count": "one_egg" }, { "cracked_state": "no_cracks", "turtle_egg_count": "two_egg" }, { "cracked_state": "no_cracks", "turtle_egg_count": "three_egg" }, { "cracked_state": "no_cracks", "turtle_egg_count": "four_egg" }],
    "minecraft:acacia_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:purpur_block": [{ "pillar_axis": "y", "chisel_type": "default" }, { "pillar_axis": "y", "chisel_type": "chiseled" }, { "pillar_axis": "y", "chisel_type": "lines" }, { "pillar_axis": "y", "chisel_type": "smooth" }, { "pillar_axis": "x", "chisel_type": "default" }, { "pillar_axis": "x", "chisel_type": "chiseled" }, { "pillar_axis": "x", "chisel_type": "lines" }, { "pillar_axis": "x", "chisel_type": "smooth" }, { "pillar_axis": "z", "chisel_type": "default" }, { "pillar_axis": "z", "chisel_type": "chiseled" }, { "pillar_axis": "z", "chisel_type": "lines" }, { "pillar_axis": "z", "chisel_type": "smooth" }, { "pillar_axis": "y", "chisel_type": "default" }, { "pillar_axis": "y", "chisel_type": "chiseled" }, { "pillar_axis": "y", "chisel_type": "lines" }, { "pillar_axis": "y", "chisel_type": "smooth" }],
    "minecraft:tnt": [{ "allow_underwater_bit": false, "explode_bit": false }, { "allow_underwater_bit": false, "explode_bit": true }, { "allow_underwater_bit": true, "explode_bit": false }, { "allow_underwater_bit": true, "explode_bit": true }],
    "minecraft:purple_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:birch_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:jungle_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:birch_wall_sign": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:trip_wire": [{ "suspended_bit": true, "powered_bit": false, "disarmed_bit": false, "attached_bit": false }, { "suspended_bit": false, "powered_bit": true, "disarmed_bit": false, "attached_bit": false }, { "suspended_bit": true, "powered_bit": false, "disarmed_bit": false, "attached_bit": false }, { "suspended_bit": true, "powered_bit": true, "disarmed_bit": false, "attached_bit": false }, { "suspended_bit": false, "powered_bit": false, "disarmed_bit": false, "attached_bit": true }, { "suspended_bit": false, "powered_bit": true, "disarmed_bit": false, "attached_bit": true }, { "suspended_bit": true, "powered_bit": false, "disarmed_bit": false, "attached_bit": true }, { "suspended_bit": true, "powered_bit": true, "disarmed_bit": false, "attached_bit": true }, { "suspended_bit": false, "powered_bit": false, "disarmed_bit": true, "attached_bit": false }, { "suspended_bit": false, "powered_bit": true, "disarmed_bit": true, "attached_bit": false }, { "suspended_bit": true, "powered_bit": false, "disarmed_bit": true, "attached_bit": false }, { "suspended_bit": true, "powered_bit": true, "disarmed_bit": true, "attached_bit": false }, { "suspended_bit": false, "powered_bit": false, "disarmed_bit": true, "attached_bit": true }, { "suspended_bit": false, "powered_bit": true, "disarmed_bit": true, "attached_bit": true }, { "suspended_bit": true, "powered_bit": false, "disarmed_bit": true, "attached_bit": true }, { "suspended_bit": true, "powered_bit": true, "disarmed_bit": true, "attached_bit": true }, { "suspended_bit": false, "powered_bit": false, "disarmed_bit": false, "attached_bit": false }],
    "minecraft:lectern": [{ "powered_bit": false, "direction": 0 }, { "powered_bit": false, "direction": 1 }, { "powered_bit": false, "direction": 2 }, { "powered_bit": false, "direction": 3 }, { "powered_bit": true, "direction": 0 }, { "powered_bit": true, "direction": 1 }, { "powered_bit": true, "direction": 2 }, { "powered_bit": true, "direction": 3 }],
    "minecraft:green_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:red_sandstone": [{ "sand_stone_type": "default" }, { "sand_stone_type": "heiroglyphs" }, { "sand_stone_type": "cut" }, { "sand_stone_type": "smooth" }],
    "minecraft:polished_diorite_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:waxed_weathered_cut_copper_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:brown_mushroom_block": [{ "huge_mushroom_bits": 0 }, { "huge_mushroom_bits": 1 }, { "huge_mushroom_bits": 2 }, { "huge_mushroom_bits": 3 }, { "huge_mushroom_bits": 4 }, { "huge_mushroom_bits": 5 }, { "huge_mushroom_bits": 6 }, { "huge_mushroom_bits": 7 }, { "huge_mushroom_bits": 8 }, { "huge_mushroom_bits": 9 }, { "huge_mushroom_bits": 10 }, { "huge_mushroom_bits": 11 }, { "huge_mushroom_bits": 12 }, { "huge_mushroom_bits": 13 }, { "huge_mushroom_bits": 14 }, { "huge_mushroom_bits": 15 }],
    "minecraft:crimson_standing_sign": [{ "ground_sign_direction": 0 }, { "ground_sign_direction": 1 }, { "ground_sign_direction": 2 }, { "ground_sign_direction": 3 }, { "ground_sign_direction": 4 }, { "ground_sign_direction": 5 }, { "ground_sign_direction": 6 }, { "ground_sign_direction": 7 }, { "ground_sign_direction": 8 }, { "ground_sign_direction": 9 }, { "ground_sign_direction": 10 }, { "ground_sign_direction": 11 }, { "ground_sign_direction": 12 }, { "ground_sign_direction": 13 }, { "ground_sign_direction": 14 }, { "ground_sign_direction": 15 }],
    "minecraft:warped_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:purple_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:flowing_lava": [{ "liquid_depth": 0 }, { "liquid_depth": 1 }, { "liquid_depth": 2 }, { "liquid_depth": 3 }, { "liquid_depth": 4 }, { "liquid_depth": 5 }, { "liquid_depth": 6 }, { "liquid_depth": 7 }, { "liquid_depth": 8 }, { "liquid_depth": 9 }, { "liquid_depth": 10 }, { "liquid_depth": 11 }, { "liquid_depth": 12 }, { "liquid_depth": 13 }, { "liquid_depth": 14 }, { "liquid_depth": 15 }],
    "minecraft:red_nether_brick_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:chain_command_block": [{ "facing_direction": "down", "conditional_bit": false }, { "facing_direction": "up", "conditional_bit": false }, { "facing_direction": "north", "conditional_bit": false }, { "facing_direction": "south", "conditional_bit": false }, { "facing_direction": "west", "conditional_bit": false }, { "facing_direction": "east", "conditional_bit": false }],
    "minecraft:double_stone_block_slab4": [{ "stone_slab_type_4": "mossy_stone_brick", "top_slot_bit": false }, { "stone_slab_type_4": "smooth_quartz", "top_slot_bit": false }, { "stone_slab_type_4": "stone", "top_slot_bit": false }, { "stone_slab_type_4": "cut_sandstone", "top_slot_bit": false }, { "stone_slab_type_4": "cut_red_sandstone", "top_slot_bit": false }],
    "minecraft:magenta_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:furnace": [{ "facing_direction": "south" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "facing_direction": "down" }],
    "minecraft:wooden_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:standing_banner": [{ "ground_sign_direction": 0 }, { "ground_sign_direction": 1 }, { "ground_sign_direction": 2 }, { "ground_sign_direction": 3 }, { "ground_sign_direction": 4 }, { "ground_sign_direction": 5 }, { "ground_sign_direction": 6 }, { "ground_sign_direction": 7 }, { "ground_sign_direction": 8 }, { "ground_sign_direction": 9 }, { "ground_sign_direction": 10 }, { "ground_sign_direction": 11 }, { "ground_sign_direction": 12 }, { "ground_sign_direction": 13 }, { "ground_sign_direction": 14 }, { "ground_sign_direction": 15 }],
    "minecraft:weathered_cut_copper_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:brick_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:prismarine": [{ "prismarine_block_type": "default" }, { "prismarine_block_type": "dark" }, { "prismarine_block_type": "bricks" }],
    "minecraft:polished_blackstone_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:sticky_piston_arm_collision": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:lime_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:double_stone_block_slab": [{ "stone_slab_type": "smooth_stone", "top_slot_bit": false }, { "stone_slab_type": "sandstone", "top_slot_bit": false }, { "stone_slab_type": "wood", "top_slot_bit": false }, { "stone_slab_type": "cobblestone", "top_slot_bit": false }, { "stone_slab_type": "brick", "top_slot_bit": false }, { "stone_slab_type": "stone_brick", "top_slot_bit": false }, { "stone_slab_type": "quartz", "top_slot_bit": false }, { "stone_slab_type": "nether_brick", "top_slot_bit": false }, { "stone_slab_type": "smooth_stone", "top_slot_bit": true }, { "stone_slab_type": "sandstone", "top_slot_bit": true }, { "stone_slab_type": "wood", "top_slot_bit": true }, { "stone_slab_type": "cobblestone", "top_slot_bit": true }, { "stone_slab_type": "brick", "top_slot_bit": true }, { "stone_slab_type": "stone_brick", "top_slot_bit": true }, { "stone_slab_type": "quartz", "top_slot_bit": true }, { "stone_slab_type": "nether_brick", "top_slot_bit": true }],
    "minecraft:cave_vines_body_with_berries": [{ "growing_plant_age": 22 }, { "growing_plant_age": 0 }],
    "minecraft:blue_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:crimson_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:stripped_birch_log": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:crimson_stem": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:end_portal_frame": [{ "end_portal_eye_bit": false, "direction": 0 }, { "end_portal_eye_bit": false, "direction": 1 }, { "end_portal_eye_bit": false, "direction": 2 }, { "end_portal_eye_bit": false, "direction": 3 }, { "end_portal_eye_bit": true, "direction": 0 }, { "end_portal_eye_bit": true, "direction": 1 }, { "end_portal_eye_bit": true, "direction": 2 }, { "end_portal_eye_bit": true, "direction": 3 }],
    "minecraft:stripped_acacia_log": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:light_gray_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:spruce_trapdoor": [{ "upside_down_bit": false, "open_bit": false, "direction": 0 }, { "upside_down_bit": false, "open_bit": false, "direction": 1 }, { "upside_down_bit": false, "open_bit": false, "direction": 2 }, { "upside_down_bit": false, "open_bit": false, "direction": 3 }, { "upside_down_bit": true, "open_bit": false, "direction": 0 }, { "upside_down_bit": true, "open_bit": false, "direction": 1 }, { "upside_down_bit": true, "open_bit": false, "direction": 2 }, { "upside_down_bit": true, "open_bit": false, "direction": 3 }, { "upside_down_bit": false, "open_bit": true, "direction": 0 }, { "upside_down_bit": false, "open_bit": true, "direction": 1 }, { "upside_down_bit": false, "open_bit": true, "direction": 2 }, { "upside_down_bit": false, "open_bit": true, "direction": 3 }, { "upside_down_bit": true, "open_bit": true, "direction": 0 }, { "upside_down_bit": true, "open_bit": true, "direction": 1 }, { "upside_down_bit": true, "open_bit": true, "direction": 2 }, { "upside_down_bit": true, "open_bit": true, "direction": 3 }],
    "minecraft:cocoa": [{ "direction": 0, "age": 0 }, { "direction": 1, "age": 0 }, { "direction": 2, "age": 0 }, { "direction": 3, "age": 0 }, { "direction": 0, "age": 1 }, { "direction": 1, "age": 1 }, { "direction": 2, "age": 1 }, { "direction": 3, "age": 1 }, { "direction": 0, "age": 2 }, { "direction": 1, "age": 2 }, { "direction": 2, "age": 2 }, { "direction": 3, "age": 2 }, { "direction": 0, "age": 0 }, { "direction": 1, "age": 0 }, { "direction": 2, "age": 0 }, { "direction": 3, "age": 0 }],
    "minecraft:hopper": [{ "toggle_bit": false, "facing_direction": "down" }, { "toggle_bit": false, "facing_direction": "up" }, { "toggle_bit": false, "facing_direction": "north" }, { "toggle_bit": false, "facing_direction": "south" }, { "toggle_bit": false, "facing_direction": "west" }, { "toggle_bit": false, "facing_direction": "east" }],
    "minecraft:black_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:log": [{ "pillar_axis": "y", "old_log_type": "oak" }, { "pillar_axis": "y", "old_log_type": "spruce" }, { "pillar_axis": "y", "old_log_type": "birch" }, { "pillar_axis": "y", "old_log_type": "jungle" }, { "pillar_axis": "x", "old_log_type": "oak" }, { "pillar_axis": "x", "old_log_type": "spruce" }, { "pillar_axis": "x", "old_log_type": "birch" }, { "pillar_axis": "x", "old_log_type": "jungle" }, { "pillar_axis": "z", "old_log_type": "oak" }, { "pillar_axis": "z", "old_log_type": "spruce" }, { "pillar_axis": "z", "old_log_type": "birch" }, { "pillar_axis": "z", "old_log_type": "jungle" }, { "pillar_axis": "y", "stripped_bit": false, "wood_type": "oak" }, { "pillar_axis": "y", "stripped_bit": false, "wood_type": "spruce" }, { "pillar_axis": "y", "stripped_bit": false, "wood_type": "birch" }, { "pillar_axis": "y", "stripped_bit": false, "wood_type": "jungle" }],
    "minecraft:planks": [{ "wood_type": "oak" }, { "wood_type": "spruce" }, { "wood_type": "birch" }, { "wood_type": "jungle" }, { "wood_type": "acacia" }, { "wood_type": "dark_oak" }],
    "minecraft:dirt": [{ "dirt_type": "normal" }, { "dirt_type": "coarse" }],
    "minecraft:end_rod": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "block_light_level": 14 }],
    "minecraft:concrete_powder": [{ "color": "white" }, { "color": "orange" }, { "color": "magenta" }, { "color": "light_blue" }, { "color": "yellow" }, { "color": "lime" }, { "color": "pink" }, { "color": "gray" }, { "color": "silver" }, { "color": "cyan" }, { "color": "purple" }, { "color": "blue" }, { "color": "brown" }, { "color": "green" }, { "color": "red" }, { "color": "black" }],
    "minecraft:sea_pickle": [{ "cluster_count": 0, "dead_bit": false }, { "cluster_count": 1, "dead_bit": false }, { "cluster_count": 2, "dead_bit": false }, { "cluster_count": 3, "dead_bit": false }, { "cluster_count": 0, "dead_bit": true }, { "cluster_count": 1, "dead_bit": true }, { "cluster_count": 2, "dead_bit": true }, { "cluster_count": 3, "dead_bit": true }],
    "minecraft:unpowered_comparator": [{ "output_lit_bit": false, "output_subtract_bit": false, "direction": 0 }, { "output_lit_bit": false, "output_subtract_bit": false, "direction": 1 }, { "output_lit_bit": false, "output_subtract_bit": false, "direction": 2 }, { "output_lit_bit": false, "output_subtract_bit": false, "direction": 3 }, { "output_lit_bit": false, "output_subtract_bit": true, "direction": 0 }, { "output_lit_bit": false, "output_subtract_bit": true, "direction": 1 }, { "output_lit_bit": false, "output_subtract_bit": true, "direction": 2 }, { "output_lit_bit": false, "output_subtract_bit": true, "direction": 3 }, { "output_lit_bit": true, "output_subtract_bit": false, "direction": 0 }, { "output_lit_bit": true, "output_subtract_bit": false, "direction": 1 }, { "output_lit_bit": true, "output_subtract_bit": false, "direction": 2 }, { "output_lit_bit": true, "output_subtract_bit": false, "direction": 3 }, { "output_lit_bit": true, "output_subtract_bit": true, "direction": 0 }, { "output_lit_bit": true, "output_subtract_bit": true, "direction": 1 }, { "output_lit_bit": true, "output_subtract_bit": true, "direction": 2 }, { "output_lit_bit": true, "output_subtract_bit": true, "direction": 3 }],
    "minecraft:polished_blackstone_pressure_plate": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:stained_hardened_clay": [{ "color": "white" }, { "color": "orange" }, { "color": "magenta" }, { "color": "light_blue" }, { "color": "yellow" }, { "color": "lime" }, { "color": "pink" }, { "color": "gray" }, { "color": "silver" }, { "color": "cyan" }, { "color": "purple" }, { "color": "blue" }, { "color": "brown" }, { "color": "green" }, { "color": "red" }, { "color": "black" }],
    "minecraft:leaves": [{ "old_leaf_type": "oak", "update_bit": false, "persistent_bit": true }, { "old_leaf_type": "spruce", "update_bit": false, "persistent_bit": true }, { "old_leaf_type": "birch", "update_bit": false, "persistent_bit": true }, { "old_leaf_type": "jungle", "update_bit": false, "persistent_bit": true }, { "old_leaf_type": "oak", "update_bit": true, "persistent_bit": true }, { "old_leaf_type": "spruce", "update_bit": true, "persistent_bit": true }, { "old_leaf_type": "birch", "update_bit": true, "persistent_bit": true }, { "old_leaf_type": "jungle", "update_bit": true, "persistent_bit": true }, { "old_leaf_type": "oak", "update_bit": false, "persistent_bit": true }, { "old_leaf_type": "spruce", "update_bit": false, "persistent_bit": true }, { "old_leaf_type": "birch", "update_bit": false, "persistent_bit": true }, { "old_leaf_type": "jungle", "update_bit": false, "persistent_bit": true }, { "old_leaf_type": "oak", "update_bit": true, "persistent_bit": true }, { "old_leaf_type": "spruce", "update_bit": true, "persistent_bit": true }, { "old_leaf_type": "birch", "update_bit": true, "persistent_bit": true }, { "old_leaf_type": "jungle", "update_bit": true, "persistent_bit": true }],
    "minecraft:campfire": [{ "extinguished": false, "direction": 0 }, { "extinguished": false, "direction": 1 }, { "extinguished": false, "direction": 2 }, { "extinguished": false, "direction": 3 }, { "extinguished": true, "direction": 0 }, { "extinguished": true, "direction": 1 }, { "extinguished": true, "direction": 2 }, { "extinguished": true, "direction": 3 }],
    "minecraft:stripped_warped_hyphae": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:jungle_standing_sign": [{ "ground_sign_direction": 0 }, { "ground_sign_direction": 1 }, { "ground_sign_direction": 2 }, { "ground_sign_direction": 3 }, { "ground_sign_direction": 4 }, { "ground_sign_direction": 5 }, { "ground_sign_direction": 6 }, { "ground_sign_direction": 7 }, { "ground_sign_direction": 8 }, { "ground_sign_direction": 9 }, { "ground_sign_direction": 10 }, { "ground_sign_direction": 11 }, { "ground_sign_direction": 12 }, { "ground_sign_direction": 13 }, { "ground_sign_direction": 14 }, { "ground_sign_direction": 15 }],
    "minecraft:acacia_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:polished_andesite_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:birch_door": [{ "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }],
    "minecraft:trapped_chest": [{ "facing_direction": "north" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "facing_direction": "down" }],
    "minecraft:white_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:bell": [{ "attachment": "standing", "toggle_bit": false, "direction": 0 }, { "attachment": "standing", "toggle_bit": false, "direction": 1 }, { "attachment": "standing", "toggle_bit": false, "direction": 2 }, { "attachment": "standing", "toggle_bit": false, "direction": 3 }, { "attachment": "hanging", "toggle_bit": false, "direction": 0 }, { "attachment": "hanging", "toggle_bit": false, "direction": 1 }, { "attachment": "hanging", "toggle_bit": false, "direction": 2 }, { "attachment": "hanging", "toggle_bit": false, "direction": 3 }, { "attachment": "side", "toggle_bit": false, "direction": 0 }, { "attachment": "side", "toggle_bit": false, "direction": 1 }, { "attachment": "side", "toggle_bit": false, "direction": 2 }, { "attachment": "side", "toggle_bit": false, "direction": 3 }, { "attachment": "multiple", "toggle_bit": false, "direction": 0 }, { "attachment": "multiple", "toggle_bit": false, "direction": 1 }, { "attachment": "multiple", "toggle_bit": false, "direction": 2 }, { "attachment": "multiple", "toggle_bit": false, "direction": 3 }, { "attachment": "standing", "toggle_bit": true, "direction": 0 }, { "attachment": "standing", "toggle_bit": true, "direction": 1 }, { "attachment": "standing", "toggle_bit": true, "direction": 2 }, { "attachment": "standing", "toggle_bit": true, "direction": 3 }, { "attachment": "hanging", "toggle_bit": true, "direction": 0 }, { "attachment": "hanging", "toggle_bit": true, "direction": 1 }, { "attachment": "hanging", "toggle_bit": true, "direction": 2 }, { "attachment": "hanging", "toggle_bit": true, "direction": 3 }, { "attachment": "side", "toggle_bit": true, "direction": 0 }, { "attachment": "side", "toggle_bit": true, "direction": 1 }, { "attachment": "side", "toggle_bit": true, "direction": 2 }, { "attachment": "side", "toggle_bit": true, "direction": 3 }, { "attachment": "multiple", "toggle_bit": true, "direction": 0 }, { "attachment": "multiple", "toggle_bit": true, "direction": 1 }, { "attachment": "multiple", "toggle_bit": true, "direction": 2 }, { "attachment": "multiple", "toggle_bit": true, "direction": 3 }],
    "minecraft:wheat": [{ "growth": 0 }, { "growth": 1 }, { "growth": 2 }, { "growth": 3 }, { "growth": 4 }, { "growth": 5 }, { "growth": 6 }, { "growth": 7 }],
    "minecraft:stripped_crimson_stem": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:pumpkin_stem": [{ "growth": 0, "facing_direction": "down" }, { "growth": 1, "facing_direction": "down" }, { "growth": 2, "facing_direction": "down" }, { "growth": 3, "facing_direction": "down" }, { "growth": 4, "facing_direction": "down" }, { "growth": 5, "facing_direction": "down" }, { "growth": 6, "facing_direction": "down" }, { "growth": 7, "facing_direction": "down" }],
    "minecraft:stone_block_slab3": [{ "stone_slab_type_3": "end_stone_brick", "top_slot_bit": false }, { "stone_slab_type_3": "smooth_red_sandstone", "top_slot_bit": false }, { "stone_slab_type_3": "polished_andesite", "top_slot_bit": false }, { "stone_slab_type_3": "andesite", "top_slot_bit": false }, { "stone_slab_type_3": "diorite", "top_slot_bit": false }, { "stone_slab_type_3": "polished_diorite", "top_slot_bit": false }, { "stone_slab_type_3": "granite", "top_slot_bit": false }, { "stone_slab_type_3": "polished_granite", "top_slot_bit": false }, { "stone_slab_type_3": "end_stone_brick", "top_slot_bit": true }, { "stone_slab_type_3": "smooth_red_sandstone", "top_slot_bit": true }, { "stone_slab_type_3": "polished_andesite", "top_slot_bit": true }, { "stone_slab_type_3": "andesite", "top_slot_bit": true }, { "stone_slab_type_3": "diorite", "top_slot_bit": true }, { "stone_slab_type_3": "polished_diorite", "top_slot_bit": true }, { "stone_slab_type_3": "granite", "top_slot_bit": true }, { "stone_slab_type_3": "polished_granite", "top_slot_bit": true }],
    "minecraft:smooth_red_sandstone_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:spruce_wall_sign": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:light_gray_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:dark_oak_pressure_plate": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:cyan_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:oxidized_cut_copper_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:golden_rail": [{ "rail_direction": 0, "rail_data_bit": false }, { "rail_direction": 1, "rail_data_bit": false }, { "rail_direction": 2, "rail_data_bit": false }, { "rail_direction": 3, "rail_data_bit": false }, { "rail_direction": 4, "rail_data_bit": false }, { "rail_direction": 5, "rail_data_bit": false }],
    "minecraft:composter": [{ "composter_fill_level": 0 }, { "composter_fill_level": 1 }, { "composter_fill_level": 2 }, { "composter_fill_level": 3 }, { "composter_fill_level": 4 }, { "composter_fill_level": 5 }, { "composter_fill_level": 6 }, { "composter_fill_level": 7 }, { "composter_fill_level": 8 }],
    "minecraft:orange_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:lit_furnace": [{ "facing_direction": "south" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "facing_direction": "down" }],
    "minecraft:reeds": [{ "age": 0 }, { "age": 1 }, { "age": 2 }, { "age": 3 }, { "age": 4 }, { "age": 5 }, { "age": 6 }, { "age": 7 }, { "age": 8 }, { "age": 9 }, { "age": 10 }, { "age": 11 }, { "age": 12 }, { "age": 13 }, { "age": 14 }, { "age": 15 }],
    "minecraft:deepslate_tile_wall": [{ "wall_connection_type_west": "tall", "wall_connection_type_south": "tall", "wall_connection_type_east": "tall", "wall_connection_type_north": "tall", "wall_post_bit": true }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false }],
    "minecraft:red_sandstone_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:skull": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:cut_copper_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:wooden_slab": [{ "wood_type": "oak", "top_slot_bit": false }, { "wood_type": "spruce", "top_slot_bit": false }, { "wood_type": "birch", "top_slot_bit": false }, { "wood_type": "jungle", "top_slot_bit": false }, { "wood_type": "acacia", "top_slot_bit": false }, { "wood_type": "dark_oak", "top_slot_bit": false }],
    "minecraft:white_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:grindstone": [{ "attachment": "standing", "direction": 0 }, { "attachment": "standing", "direction": 1 }, { "attachment": "standing", "direction": 2 }, { "attachment": "standing", "direction": 3 }, { "attachment": "hanging", "direction": 0 }, { "attachment": "hanging", "direction": 1 }, { "attachment": "hanging", "direction": 2 }, { "attachment": "hanging", "direction": 3 }, { "attachment": "side", "direction": 0 }, { "attachment": "side", "direction": 1 }, { "attachment": "side", "direction": 2 }, { "attachment": "side", "direction": 3 }, { "attachment": "multiple", "direction": 0 }, { "attachment": "multiple", "direction": 1 }, { "attachment": "multiple", "direction": 2 }, { "attachment": "multiple", "direction": 3 }],
    "minecraft:coral_block": [{ "coral_color": "blue", "dead_bit": false }, { "coral_color": "pink", "dead_bit": false }, { "coral_color": "purple", "dead_bit": false }, { "coral_color": "red", "dead_bit": false }, { "coral_color": "yellow", "dead_bit": false }],
    "minecraft:lever": [{ "lever_direction": "down_east_west", "open_bit": false }, { "lever_direction": "east", "open_bit": false }, { "lever_direction": "west", "open_bit": false }, { "lever_direction": "south", "open_bit": false }, { "lever_direction": "north", "open_bit": false }, { "lever_direction": "up_north_south", "open_bit": false }, { "lever_direction": "up_east_west", "open_bit": false }, { "lever_direction": "down_north_south", "open_bit": false }, { "lever_direction": "down_east_west", "open_bit": true }, { "lever_direction": "east", "open_bit": true }, { "lever_direction": "west", "open_bit": true }, { "lever_direction": "south", "open_bit": true }, { "lever_direction": "north", "open_bit": true }, { "lever_direction": "up_north_south", "open_bit": true }, { "lever_direction": "up_east_west", "open_bit": true }, { "lever_direction": "down_north_south", "open_bit": true }],
    "minecraft:mangrove_door": [{ "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }],
    "minecraft:cave_vines": [{ "growing_plant_age": 25 }, { "growing_plant_age": 0 }],
    "minecraft:mangrove_wall_sign": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:water": [{ "liquid_depth": 0 }, { "liquid_depth": 1 }, { "liquid_depth": 2 }, { "liquid_depth": 3 }, { "liquid_depth": 4 }, { "liquid_depth": 5 }, { "liquid_depth": 6 }, { "liquid_depth": 7 }, { "liquid_depth": 8 }, { "liquid_depth": 9 }, { "liquid_depth": 10 }, { "liquid_depth": 11 }, { "liquid_depth": 12 }, { "liquid_depth": 13 }, { "liquid_depth": 14 }, { "liquid_depth": 15 }],
    "minecraft:beetroot": [{ "growth": 0 }, { "growth": 1 }, { "growth": 2 }, { "growth": 3 }, { "growth": 4 }, { "growth": 5 }, { "growth": 6 }, { "growth": 7 }],
    "minecraft:portal": [{ "portal_axis": "unknown" }, { "portal_axis": "x" }, { "portal_axis": "z" }],
    "minecraft:jungle_trapdoor": [{ "upside_down_bit": false, "open_bit": false, "direction": 0 }, { "upside_down_bit": false, "open_bit": false, "direction": 1 }, { "upside_down_bit": false, "open_bit": false, "direction": 2 }, { "upside_down_bit": false, "open_bit": false, "direction": 3 }, { "upside_down_bit": true, "open_bit": false, "direction": 0 }, { "upside_down_bit": true, "open_bit": false, "direction": 1 }, { "upside_down_bit": true, "open_bit": false, "direction": 2 }, { "upside_down_bit": true, "open_bit": false, "direction": 3 }, { "upside_down_bit": false, "open_bit": true, "direction": 0 }, { "upside_down_bit": false, "open_bit": true, "direction": 1 }, { "upside_down_bit": false, "open_bit": true, "direction": 2 }, { "upside_down_bit": false, "open_bit": true, "direction": 3 }, { "upside_down_bit": true, "open_bit": true, "direction": 0 }, { "upside_down_bit": true, "open_bit": true, "direction": 1 }, { "upside_down_bit": true, "open_bit": true, "direction": 2 }, { "upside_down_bit": true, "open_bit": true, "direction": 3 }],
    "minecraft:jungle_fence_gate": [{ "open_bit": false, "in_wall_bit": false, "direction": 0 }, { "open_bit": false, "in_wall_bit": false, "direction": 1 }, { "open_bit": false, "in_wall_bit": false, "direction": 2 }, { "open_bit": false, "in_wall_bit": false, "direction": 3 }, { "open_bit": true, "in_wall_bit": false, "direction": 0 }, { "open_bit": true, "in_wall_bit": false, "direction": 1 }, { "open_bit": true, "in_wall_bit": false, "direction": 2 }, { "open_bit": true, "in_wall_bit": false, "direction": 3 }, { "open_bit": false, "in_wall_bit": true, "direction": 0 }, { "open_bit": false, "in_wall_bit": true, "direction": 1 }, { "open_bit": false, "in_wall_bit": true, "direction": 2 }, { "open_bit": false, "in_wall_bit": true, "direction": 3 }, { "open_bit": true, "in_wall_bit": true, "direction": 0 }, { "open_bit": true, "in_wall_bit": true, "direction": 1 }, { "open_bit": true, "in_wall_bit": true, "direction": 2 }, { "open_bit": true, "in_wall_bit": true, "direction": 3 }],
    "minecraft:frosted_ice": [{ "age": 0 }, { "age": 1 }, { "age": 2 }, { "age": 3 }],
    "minecraft:yellow_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:lime_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:pumpkin": [{ "direction": 0 }, { "direction": 1 }, { "direction": 2 }, { "direction": 3 }],
    "minecraft:coral_fan_hang": [{ "coral_hang_type_bit": false, "coral_direction": 0, "dead_bit": false }, { "coral_hang_type_bit": true, "coral_direction": 0, "dead_bit": false }, { "coral_hang_type_bit": false, "coral_direction": 0, "dead_bit": true }, { "coral_hang_type_bit": true, "coral_direction": 0, "dead_bit": true }, { "coral_hang_type_bit": false, "coral_direction": 1, "dead_bit": false }],
    "minecraft:mossy_stone_brick_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:mangrove_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:chorus_flower": [{ "age": 0 }, { "age": 1 }, { "age": 2 }, { "age": 3 }, { "age": 4 }, { "age": 5 }],
    "minecraft:dark_oak_trapdoor": [{ "upside_down_bit": false, "open_bit": false, "direction": 0 }, { "upside_down_bit": false, "open_bit": false, "direction": 1 }, { "upside_down_bit": false, "open_bit": false, "direction": 2 }, { "upside_down_bit": false, "open_bit": false, "direction": 3 }, { "upside_down_bit": true, "open_bit": false, "direction": 0 }, { "upside_down_bit": true, "open_bit": false, "direction": 1 }, { "upside_down_bit": true, "open_bit": false, "direction": 2 }, { "upside_down_bit": true, "open_bit": false, "direction": 3 }, { "upside_down_bit": false, "open_bit": true, "direction": 0 }, { "upside_down_bit": false, "open_bit": true, "direction": 1 }, { "upside_down_bit": false, "open_bit": true, "direction": 2 }, { "upside_down_bit": false, "open_bit": true, "direction": 3 }, { "upside_down_bit": true, "open_bit": true, "direction": 0 }, { "upside_down_bit": true, "open_bit": true, "direction": 1 }, { "upside_down_bit": true, "open_bit": true, "direction": 2 }, { "upside_down_bit": true, "open_bit": true, "direction": 3 }],
    "minecraft:crimson_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:weeping_vines": [{ "weeping_vines_age": 6 }, { "weeping_vines_age": 1 }, { "weeping_vines_age": 2 }, { "weeping_vines_age": 3 }, { "weeping_vines_age": 4 }, { "weeping_vines_age": 5 }, { "weeping_vines_age": 6 }, { "weeping_vines_age": 7 }, { "weeping_vines_age": 8 }, { "weeping_vines_age": 9 }, { "weeping_vines_age": 10 }, { "weeping_vines_age": 11 }, { "weeping_vines_age": 12 }, { "weeping_vines_age": 13 }, { "weeping_vines_age": 14 }, { "weeping_vines_age": 15 }, { "weeping_vines_age": 16 }, { "weeping_vines_age": 17 }, { "weeping_vines_age": 18 }, { "weeping_vines_age": 19 }, { "weeping_vines_age": 20 }, { "weeping_vines_age": 21 }, { "weeping_vines_age": 22 }, { "weeping_vines_age": 23 }, { "weeping_vines_age": 24 }, { "weeping_vines_age": 25 }, { "weeping_vines_age": 0 }],
    "minecraft:yellow_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:prismarine_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:diorite_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:acacia_pressure_plate": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:stone_block_slab2": [{ "stone_slab_type_2": "red_sandstone", "top_slot_bit": false }, { "stone_slab_type_2": "purpur", "top_slot_bit": false }, { "stone_slab_type_2": "prismarine_rough", "top_slot_bit": false }, { "stone_slab_type_2": "prismarine_dark", "top_slot_bit": false }, { "stone_slab_type_2": "prismarine_brick", "top_slot_bit": false }, { "stone_slab_type_2": "mossy_cobblestone", "top_slot_bit": false }, { "stone_slab_type_2": "smooth_sandstone", "top_slot_bit": false }, { "stone_slab_type_2": "red_nether_brick", "top_slot_bit": false }, { "stone_slab_type_2": "red_sandstone", "top_slot_bit": true }, { "stone_slab_type_2": "purpur", "top_slot_bit": true }, { "stone_slab_type_2": "prismarine_rough", "top_slot_bit": true }, { "stone_slab_type_2": "prismarine_dark", "top_slot_bit": true }, { "stone_slab_type_2": "prismarine_brick", "top_slot_bit": true }, { "stone_slab_type_2": "mossy_cobblestone", "top_slot_bit": true }, { "stone_slab_type_2": "smooth_sandstone", "top_slot_bit": true }, { "stone_slab_type_2": "red_nether_brick", "top_slot_bit": true }],
    "minecraft:stained_glass_pane": [{ "color": "white" }, { "color": "orange" }, { "color": "magenta" }, { "color": "light_blue" }, { "color": "yellow" }, { "color": "lime" }, { "color": "pink" }, { "color": "gray" }, { "color": "silver" }, { "color": "cyan" }, { "color": "purple" }, { "color": "blue" }, { "color": "brown" }, { "color": "green" }, { "color": "red" }, { "color": "black" }],
    "minecraft:jungle_wall_sign": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:ender_chest": [{ "facing_direction": "north" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "facing_direction": "down" }],
    "minecraft:light_blue_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:warped_trapdoor": [{ "upside_down_bit": false, "open_bit": false, "direction": 0 }, { "upside_down_bit": false, "open_bit": false, "direction": 1 }, { "upside_down_bit": false, "open_bit": false, "direction": 2 }, { "upside_down_bit": false, "open_bit": false, "direction": 3 }, { "upside_down_bit": true, "open_bit": false, "direction": 0 }, { "upside_down_bit": true, "open_bit": false, "direction": 1 }, { "upside_down_bit": true, "open_bit": false, "direction": 2 }, { "upside_down_bit": true, "open_bit": false, "direction": 3 }, { "upside_down_bit": false, "open_bit": true, "direction": 0 }, { "upside_down_bit": false, "open_bit": true, "direction": 1 }, { "upside_down_bit": false, "open_bit": true, "direction": 2 }, { "upside_down_bit": false, "open_bit": true, "direction": 3 }, { "upside_down_bit": true, "open_bit": true, "direction": 0 }, { "upside_down_bit": true, "open_bit": true, "direction": 1 }, { "upside_down_bit": true, "open_bit": true, "direction": 2 }, { "upside_down_bit": true, "open_bit": true, "direction": 3 }],
    "minecraft:end_brick_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:basalt": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:dark_oak_door": [{ "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }],
    "minecraft:flowing_water": [{ "liquid_depth": 0 }, {}, { "liquid_depth": 2 }, { "liquid_depth": 3 }, { "liquid_depth": 4 }, {}, { "liquid_depth": 6 }, { "liquid_depth": 7 }, {}, { "liquid_depth": 9 }, {}, { "liquid_depth": 11 }, { "liquid_depth": 12 }, { "liquid_depth": 13 }, { "liquid_depth": 14 }, { "liquid_depth": 15 }],
    "minecraft:sticky_piston": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:double_stone_block_slab2": [{ "stone_slab_type_2": "red_sandstone", "top_slot_bit": false }, { "stone_slab_type_2": "purpur", "top_slot_bit": false }, { "stone_slab_type_2": "prismarine_rough", "top_slot_bit": false }, { "stone_slab_type_2": "prismarine_dark", "top_slot_bit": false }, { "stone_slab_type_2": "prismarine_brick", "top_slot_bit": false }, { "stone_slab_type_2": "mossy_cobblestone", "top_slot_bit": false }, { "stone_slab_type_2": "smooth_sandstone", "top_slot_bit": false }, { "stone_slab_type_2": "red_nether_brick", "top_slot_bit": false }, { "stone_slab_type_2": "red_sandstone", "top_slot_bit": true }, { "stone_slab_type_2": "purpur", "top_slot_bit": true }, { "stone_slab_type_2": "prismarine_rough", "top_slot_bit": true }, { "stone_slab_type_2": "prismarine_dark", "top_slot_bit": true }, { "stone_slab_type_2": "prismarine_brick", "top_slot_bit": true }, { "stone_slab_type_2": "mossy_cobblestone", "top_slot_bit": true }, { "stone_slab_type_2": "smooth_sandstone", "top_slot_bit": true }, { "stone_slab_type_2": "red_nether_brick", "top_slot_bit": true }],
    "minecraft:crimson_fence_gate": [{ "open_bit": false, "in_wall_bit": false, "direction": 0 }, { "open_bit": false, "in_wall_bit": false, "direction": 1 }, { "open_bit": false, "in_wall_bit": false, "direction": 2 }, { "open_bit": false, "in_wall_bit": false, "direction": 3 }, { "open_bit": true, "in_wall_bit": false, "direction": 0 }, { "open_bit": true, "in_wall_bit": false, "direction": 1 }, { "open_bit": true, "in_wall_bit": false, "direction": 2 }, { "open_bit": true, "in_wall_bit": false, "direction": 3 }, { "open_bit": false, "in_wall_bit": true, "direction": 0 }, { "open_bit": false, "in_wall_bit": true, "direction": 1 }, { "open_bit": false, "in_wall_bit": true, "direction": 2 }, { "open_bit": false, "in_wall_bit": true, "direction": 3 }, { "open_bit": true, "in_wall_bit": true, "direction": 0 }, { "open_bit": true, "in_wall_bit": true, "direction": 1 }, { "open_bit": true, "in_wall_bit": true, "direction": 2 }, { "open_bit": true, "in_wall_bit": true, "direction": 3 }],
    "minecraft:mangrove_fence_gate": [{ "open_bit": false, "in_wall_bit": false, "direction": 0 }, { "open_bit": false, "in_wall_bit": false, "direction": 1 }, { "open_bit": false, "in_wall_bit": false, "direction": 2 }, { "open_bit": false, "in_wall_bit": false, "direction": 3 }, { "open_bit": true, "in_wall_bit": false, "direction": 0 }, { "open_bit": true, "in_wall_bit": false, "direction": 1 }, { "open_bit": true, "in_wall_bit": false, "direction": 2 }, { "open_bit": true, "in_wall_bit": false, "direction": 3 }, { "open_bit": false, "in_wall_bit": true, "direction": 0 }, { "open_bit": false, "in_wall_bit": true, "direction": 1 }, { "open_bit": false, "in_wall_bit": true, "direction": 2 }, { "open_bit": false, "in_wall_bit": true, "direction": 3 }, { "open_bit": true, "in_wall_bit": true, "direction": 0 }, { "open_bit": true, "in_wall_bit": true, "direction": 1 }, { "open_bit": true, "in_wall_bit": true, "direction": 2 }, { "open_bit": true, "in_wall_bit": true, "direction": 3 }],
    "minecraft:sand": [{ "sand_type": "normal" }, { "sand_type": "red" }],
    "minecraft:stripped_mangrove_wood": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:polished_blackstone_brick_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:structure_void": [{ "structure_void_type": "void" }, { "structure_void_type": "air" }],
    "minecraft:mangrove_trapdoor": [{ "upside_down_bit": false, "open_bit": false, "direction": 0 }, { "upside_down_bit": false, "open_bit": false, "direction": 1 }, { "upside_down_bit": false, "open_bit": false, "direction": 2 }, { "upside_down_bit": false, "open_bit": false, "direction": 3 }, { "upside_down_bit": true, "open_bit": false, "direction": 0 }, { "upside_down_bit": true, "open_bit": false, "direction": 1 }, { "upside_down_bit": true, "open_bit": false, "direction": 2 }, { "upside_down_bit": true, "open_bit": false, "direction": 3 }, { "upside_down_bit": false, "open_bit": true, "direction": 0 }, { "upside_down_bit": false, "open_bit": true, "direction": 1 }, { "upside_down_bit": false, "open_bit": true, "direction": 2 }, { "upside_down_bit": false, "open_bit": true, "direction": 3 }, { "upside_down_bit": true, "open_bit": true, "direction": 0 }, { "upside_down_bit": true, "open_bit": true, "direction": 1 }, { "upside_down_bit": true, "open_bit": true, "direction": 2 }, { "upside_down_bit": true, "open_bit": true, "direction": 3 }],
    "minecraft:polished_deepslate_wall": [{ "wall_connection_type_west": "tall", "wall_connection_type_south": "tall", "wall_connection_type_east": "tall", "wall_connection_type_north": "tall", "wall_post_bit": true }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false }],
    "minecraft:spruce_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:light_blue_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:nether_wart": [{ "age": 0 }, { "age": 1 }, { "age": 2 }, { "age": 3 }],
    "minecraft:mangrove_wood": [{ "pillar_axis": "y", "stripped_bit": false }, { "pillar_axis": "x", "stripped_bit": false }, { "pillar_axis": "z", "stripped_bit": false }],
    "minecraft:coral_fan_hang3": [{ "coral_hang_type_bit": false, "coral_direction": 0, "dead_bit": false }, { "coral_hang_type_bit": true, "coral_direction": 0, "dead_bit": false }, { "coral_hang_type_bit": false, "coral_direction": 0, "dead_bit": true }, { "coral_hang_type_bit": true, "coral_direction": 0, "dead_bit": true }, { "coral_hang_type_bit": false, "coral_direction": 1, "dead_bit": false }],
    "minecraft:red_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:coral_fan_hang2": [{ "coral_hang_type_bit": false, "coral_direction": 0, "dead_bit": false }, { "coral_hang_type_bit": true, "coral_direction": 0, "dead_bit": false }, { "coral_hang_type_bit": false, "coral_direction": 0, "dead_bit": true }, { "coral_hang_type_bit": true, "coral_direction": 0, "dead_bit": true }, { "coral_hang_type_bit": false, "coral_direction": 1, "dead_bit": false }],
    "minecraft:soul_lantern": [{ "hanging": false }, { "hanging": true }],
    "minecraft:monster_egg": [{ "monster_egg_stone_type": "stone" }, { "monster_egg_stone_type": "cobblestone" }, { "monster_egg_stone_type": "stone_brick" }, { "monster_egg_stone_type": "mossy_stone_brick" }, { "monster_egg_stone_type": "cracked_stone_brick" }, { "monster_egg_stone_type": "chiseled_stone_brick" }],
    "minecraft:detector_rail": [{ "rail_direction": 0, "rail_data_bit": false }, { "rail_direction": 1, "rail_data_bit": false }, { "rail_direction": 2, "rail_data_bit": false }, { "rail_direction": 3, "rail_data_bit": false }, { "rail_direction": 4, "rail_data_bit": false }, { "rail_direction": 5, "rail_data_bit": false }],
    "minecraft:mud_brick_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:warped_standing_sign": [{ "ground_sign_direction": 0 }, { "ground_sign_direction": 1 }, { "ground_sign_direction": 2 }, { "ground_sign_direction": 3 }, { "ground_sign_direction": 4 }, { "ground_sign_direction": 5 }, { "ground_sign_direction": 6 }, { "ground_sign_direction": 7 }, { "ground_sign_direction": 8 }, { "ground_sign_direction": 9 }, { "ground_sign_direction": 10 }, { "ground_sign_direction": 11 }, { "ground_sign_direction": 12 }, { "ground_sign_direction": 13 }, { "ground_sign_direction": 14 }, { "ground_sign_direction": 15 }],
    "minecraft:polished_blackstone_wall": [{ "wall_connection_type_west": "tall", "wall_connection_type_south": "tall", "wall_connection_type_east": "tall", "wall_connection_type_north": "tall", "wall_post_bit": true }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false }],
    "minecraft:log2": [{ "pillar_axis": "y", "new_log_type": "acacia" }, { "pillar_axis": "y", "new_log_type": "dark_oak" }],
    "minecraft:mossy_cobblestone_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:brown_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:warped_fence_gate": [{ "open_bit": false, "in_wall_bit": false, "direction": 0 }, { "open_bit": false, "in_wall_bit": false, "direction": 1 }, { "open_bit": false, "in_wall_bit": false, "direction": 2 }, { "open_bit": false, "in_wall_bit": false, "direction": 3 }, { "open_bit": true, "in_wall_bit": false, "direction": 0 }, { "open_bit": true, "in_wall_bit": false, "direction": 1 }, { "open_bit": true, "in_wall_bit": false, "direction": 2 }, { "open_bit": true, "in_wall_bit": false, "direction": 3 }, { "open_bit": false, "in_wall_bit": true, "direction": 0 }, { "open_bit": false, "in_wall_bit": true, "direction": 1 }, { "open_bit": false, "in_wall_bit": true, "direction": 2 }, { "open_bit": false, "in_wall_bit": true, "direction": 3 }, { "open_bit": true, "in_wall_bit": true, "direction": 0 }, { "open_bit": true, "in_wall_bit": true, "direction": 1 }, { "open_bit": true, "in_wall_bit": true, "direction": 2 }, { "open_bit": true, "in_wall_bit": true, "direction": 3 }],
    "minecraft:pink_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:carrots": [{ "growth": 0 }, { "growth": 1 }, { "growth": 2 }, { "growth": 3 }, { "growth": 4 }, { "growth": 5 }, { "growth": 6 }, { "growth": 7 }],
    "minecraft:crimson_trapdoor": [{ "upside_down_bit": false, "open_bit": false, "direction": 0 }, { "upside_down_bit": false, "open_bit": false, "direction": 1 }, { "upside_down_bit": false, "open_bit": false, "direction": 2 }, { "upside_down_bit": false, "open_bit": false, "direction": 3 }, { "upside_down_bit": true, "open_bit": false, "direction": 0 }, { "upside_down_bit": true, "open_bit": false, "direction": 1 }, { "upside_down_bit": true, "open_bit": false, "direction": 2 }, { "upside_down_bit": true, "open_bit": false, "direction": 3 }, { "upside_down_bit": false, "open_bit": true, "direction": 0 }, { "upside_down_bit": false, "open_bit": true, "direction": 1 }, { "upside_down_bit": false, "open_bit": true, "direction": 2 }, { "upside_down_bit": false, "open_bit": true, "direction": 3 }, { "upside_down_bit": true, "open_bit": true, "direction": 0 }, { "upside_down_bit": true, "open_bit": true, "direction": 1 }, { "upside_down_bit": true, "open_bit": true, "direction": 2 }, { "upside_down_bit": true, "open_bit": true, "direction": 3 }],
    "minecraft:bedrock": [{ "infiniburn_bit": false }, { "infiniburn_bit": true }],
    "minecraft:farmland": [{ "moisturized_amount": 0 }, { "moisturized_amount": 1 }, { "moisturized_amount": 2 }, { "moisturized_amount": 3 }, { "moisturized_amount": 4 }, { "moisturized_amount": 5 }, { "moisturized_amount": 6 }, { "moisturized_amount": 7 }],
    "minecraft:stone_block_slab": [{ "stone_slab_type": "smooth_stone", "top_slot_bit": false }, { "stone_slab_type": "sandstone", "top_slot_bit": false }, { "stone_slab_type": "wood", "top_slot_bit": false }, { "stone_slab_type": "cobblestone", "top_slot_bit": false }, { "stone_slab_type": "brick", "top_slot_bit": false }, { "stone_slab_type": "stone_brick", "top_slot_bit": false }, { "stone_slab_type": "quartz", "top_slot_bit": false }, { "stone_slab_type": "nether_brick", "top_slot_bit": false }, { "stone_slab_type": "smooth_stone", "top_slot_bit": true }, { "stone_slab_type": "sandstone", "top_slot_bit": true }, { "stone_slab_type": "wood", "top_slot_bit": true }, { "stone_slab_type": "cobblestone", "top_slot_bit": true }, { "stone_slab_type": "brick", "top_slot_bit": true }, { "stone_slab_type": "stone_brick", "top_slot_bit": true }, { "stone_slab_type": "quartz", "top_slot_bit": true }, { "stone_slab_type": "nether_brick", "top_slot_bit": true }],
    "minecraft:anvil": [{ "direction": 0, "damage": "undamaged" }, { "direction": 1, "damage": "undamaged" }, { "direction": 2, "damage": "undamaged" }, { "direction": 3, "damage": "undamaged" }, { "direction": 0, "damage": "slightly_damaged" }, { "direction": 1, "damage": "slightly_damaged" }, { "direction": 2, "damage": "slightly_damaged" }, { "direction": 3, "damage": "slightly_damaged" }, { "direction": 0, "damage": "very_damaged" }, { "direction": 1, "damage": "very_damaged" }, { "direction": 2, "damage": "very_damaged" }, { "direction": 3, "damage": "very_damaged" }, { "direction": 0, "damage": "broken" }, { "direction": 1, "damage": "broken" }, { "direction": 2, "damage": "broken" }, { "direction": 3, "damage": "broken" }],
    "minecraft:blackstone_wall": [{ "wall_connection_type_west": "tall", "wall_connection_type_south": "tall", "wall_connection_type_east": "tall", "wall_connection_type_north": "tall", "wall_post_bit": true }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false }],
    "minecraft:stonebrick": [{ "stone_brick_type": "default" }, { "stone_brick_type": "mossy" }, { "stone_brick_type": "cracked" }, { "stone_brick_type": "chiseled" }, { "stone_brick_type": "smooth" }],
    "minecraft:rail": [{ "rail_direction": 0 }, { "rail_direction": 1 }, { "rail_direction": 2 }, { "rail_direction": 3 }, { "rail_direction": 4 }, { "rail_direction": 5 }, { "rail_direction": 6 }, { "rail_direction": 7 }, { "rail_direction": 8 }, { "rail_direction": 9 }],
    "minecraft:exposed_cut_copper_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:quartz_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:mangrove_standing_sign": [{ "ground_sign_direction": 0 }, { "ground_sign_direction": 1 }, { "ground_sign_direction": 2 }, { "ground_sign_direction": 3 }, { "ground_sign_direction": 4 }, { "ground_sign_direction": 5 }, { "ground_sign_direction": 6 }, { "ground_sign_direction": 7 }, { "ground_sign_direction": 8 }, { "ground_sign_direction": 9 }, { "ground_sign_direction": 10 }, { "ground_sign_direction": 11 }, { "ground_sign_direction": 12 }, { "ground_sign_direction": 13 }, { "ground_sign_direction": 14 }, { "ground_sign_direction": 15 }],
    "minecraft:polished_blackstone_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:spruce_door": [{ "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }],
    "minecraft:dark_oak_fence_gate": [{ "open_bit": false, "in_wall_bit": false, "direction": 0 }, { "open_bit": false, "in_wall_bit": false, "direction": 1 }, { "open_bit": false, "in_wall_bit": false, "direction": 2 }, { "open_bit": false, "in_wall_bit": false, "direction": 3 }, { "open_bit": true, "in_wall_bit": false, "direction": 0 }, { "open_bit": true, "in_wall_bit": false, "direction": 1 }, { "open_bit": true, "in_wall_bit": false, "direction": 2 }, { "open_bit": true, "in_wall_bit": false, "direction": 3 }, { "open_bit": false, "in_wall_bit": true, "direction": 0 }, { "open_bit": false, "in_wall_bit": true, "direction": 1 }, { "open_bit": false, "in_wall_bit": true, "direction": 2 }, { "open_bit": false, "in_wall_bit": true, "direction": 3 }, { "open_bit": true, "in_wall_bit": true, "direction": 0 }, { "open_bit": true, "in_wall_bit": true, "direction": 1 }, { "open_bit": true, "in_wall_bit": true, "direction": 2 }, { "open_bit": true, "in_wall_bit": true, "direction": 3 }],
    "minecraft:sandstone": [{ "sand_stone_type": "default" }, { "sand_stone_type": "heiroglyphs" }, { "sand_stone_type": "cut" }, { "sand_stone_type": "smooth" }],
    "minecraft:bamboo": [{ "bamboo_leaf_size": "no_leaves", "bamboo_stalk_thickness": "thin", "age_bit": false }, { "bamboo_leaf_size": "no_leaves", "bamboo_stalk_thickness": "thick", "age_bit": false }, { "bamboo_leaf_size": "small_leaves", "bamboo_stalk_thickness": "thin", "age_bit": false }, { "bamboo_leaf_size": "small_leaves", "bamboo_stalk_thickness": "thick", "age_bit": false }, { "bamboo_leaf_size": "large_leaves", "bamboo_stalk_thickness": "thin", "age_bit": false }, { "bamboo_leaf_size": "large_leaves", "bamboo_stalk_thickness": "thick", "age_bit": false }],
    "minecraft:stripped_warped_stem": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:birch_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:acacia_trapdoor": [{ "upside_down_bit": false, "open_bit": false, "direction": 0 }, { "upside_down_bit": false, "open_bit": false, "direction": 1 }, { "upside_down_bit": false, "open_bit": false, "direction": 2 }, { "upside_down_bit": false, "open_bit": false, "direction": 3 }, { "upside_down_bit": true, "open_bit": false, "direction": 0 }, { "upside_down_bit": true, "open_bit": false, "direction": 1 }, { "upside_down_bit": true, "open_bit": false, "direction": 2 }, { "upside_down_bit": true, "open_bit": false, "direction": 3 }, { "upside_down_bit": false, "open_bit": true, "direction": 0 }, { "upside_down_bit": false, "open_bit": true, "direction": 1 }, { "upside_down_bit": false, "open_bit": true, "direction": 2 }, { "upside_down_bit": false, "open_bit": true, "direction": 3 }, { "upside_down_bit": true, "open_bit": true, "direction": 0 }, { "upside_down_bit": true, "open_bit": true, "direction": 1 }, { "upside_down_bit": true, "open_bit": true, "direction": 2 }, { "upside_down_bit": true, "open_bit": true, "direction": 3 }],
    "minecraft:iron_door": [{ "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }],
    "minecraft:warped_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:double_wooden_slab": [{ "wood_type": "oak", "top_slot_bit": false }, { "wood_type": "spruce", "top_slot_bit": false }, { "wood_type": "birch", "top_slot_bit": false }, { "wood_type": "jungle", "top_slot_bit": false }, { "wood_type": "acacia", "top_slot_bit": false }, { "wood_type": "dark_oak", "top_slot_bit": false }],
    "minecraft:pink_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:repeating_command_block": [{ "facing_direction": "down", "conditional_bit": false }, { "facing_direction": "up", "conditional_bit": false }, { "facing_direction": "north", "conditional_bit": false }, { "facing_direction": "south", "conditional_bit": false }, { "facing_direction": "west", "conditional_bit": false }, { "facing_direction": "east", "conditional_bit": false }],
    "minecraft:brown_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:stone_pressure_plate": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:silver_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:smooth_sandstone_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:redstone_wire": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:sandstone_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:redstone_torch": [{ "torch_facing_direction": "unknown" }, { "torch_facing_direction": "west" }, { "torch_facing_direction": "east" }, { "torch_facing_direction": "north" }, { "torch_facing_direction": "south" }, { "torch_facing_direction": "top" }],
    "minecraft:pink_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:warped_hyphae": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:brown_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:granite_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:observer": [{ "powered_bit": false, "facing_direction": "down" }, { "powered_bit": false, "facing_direction": "up" }, { "powered_bit": false, "facing_direction": "north" }, { "powered_bit": false, "facing_direction": "south" }, { "powered_bit": false, "facing_direction": "west" }, { "powered_bit": false, "facing_direction": "east" }],
    "minecraft:red_flower": [{ "flower_type": "poppy" }, { "flower_type": "orchid" }, { "flower_type": "allium" }, { "flower_type": "houstonia" }, { "flower_type": "tulip_red" }, { "flower_type": "tulip_orange" }, { "flower_type": "tulip_white" }, { "flower_type": "tulip_pink" }, { "flower_type": "oxeye" }, { "flower_type": "cornflower" }, { "flower_type": "lily_of_the_valley" }],
    "minecraft:spruce_fence_gate": [{ "open_bit": false, "in_wall_bit": false, "direction": 0 }, { "open_bit": false, "in_wall_bit": false, "direction": 1 }, { "open_bit": false, "in_wall_bit": false, "direction": 2 }, { "open_bit": false, "in_wall_bit": false, "direction": 3 }, { "open_bit": true, "in_wall_bit": false, "direction": 0 }, { "open_bit": true, "in_wall_bit": false, "direction": 1 }, { "open_bit": true, "in_wall_bit": false, "direction": 2 }, { "open_bit": true, "in_wall_bit": false, "direction": 3 }, { "open_bit": false, "in_wall_bit": true, "direction": 0 }, { "open_bit": false, "in_wall_bit": true, "direction": 1 }, { "open_bit": false, "in_wall_bit": true, "direction": 2 }, { "open_bit": false, "in_wall_bit": true, "direction": 3 }, { "open_bit": true, "in_wall_bit": true, "direction": 0 }, { "open_bit": true, "in_wall_bit": true, "direction": 1 }, { "open_bit": true, "in_wall_bit": true, "direction": 2 }, { "open_bit": true, "in_wall_bit": true, "direction": 3 }],
    "minecraft:mangrove_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:wood": [{ "pillar_axis": "y", "stripped_bit": false, "wood_type": "oak" }, { "pillar_axis": "y", "stripped_bit": false, "wood_type": "spruce" }, { "pillar_axis": "y", "stripped_bit": false, "wood_type": "birch" }, { "pillar_axis": "y", "stripped_bit": false, "wood_type": "jungle" }, { "pillar_axis": "y", "stripped_bit": false, "wood_type": "acacia" }, { "pillar_axis": "y", "stripped_bit": false, "wood_type": "dark_oak" }],
    "minecraft:light_weighted_pressure_plate": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:stripped_spruce_log": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:deepslate_brick_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:orange_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:jungle_pressure_plate": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:acacia_door": [{ "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }],
    "minecraft:warped_stem": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:tallgrass": [{ "tall_grass_type": "default" }, { "tall_grass_type": "tall" }, { "tall_grass_type": "fern" }, { "tall_grass_type": "snow" }],
    "minecraft:blackstone_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:mud_brick_wall": [{ "wall_connection_type_west": "tall", "wall_connection_type_south": "tall", "wall_connection_type_east": "tall", "wall_connection_type_north": "tall", "wall_post_bit": true }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false }],
    "minecraft:magenta_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:wall_sign": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:birch_fence_gate": [{ "open_bit": false, "in_wall_bit": false, "direction": 0 }, { "open_bit": false, "in_wall_bit": false, "direction": 1 }, { "open_bit": false, "in_wall_bit": false, "direction": 2 }, { "open_bit": false, "in_wall_bit": false, "direction": 3 }, { "open_bit": true, "in_wall_bit": false, "direction": 0 }, { "open_bit": true, "in_wall_bit": false, "direction": 1 }, { "open_bit": true, "in_wall_bit": false, "direction": 2 }, { "open_bit": true, "in_wall_bit": false, "direction": 3 }, { "open_bit": false, "in_wall_bit": true, "direction": 0 }, { "open_bit": false, "in_wall_bit": true, "direction": 1 }, { "open_bit": false, "in_wall_bit": true, "direction": 2 }, { "open_bit": false, "in_wall_bit": true, "direction": 3 }, { "open_bit": true, "in_wall_bit": true, "direction": 0 }, { "open_bit": true, "in_wall_bit": true, "direction": 1 }, { "open_bit": true, "in_wall_bit": true, "direction": 2 }, { "open_bit": true, "in_wall_bit": true, "direction": 3 }],
    "minecraft:piston": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:lit_pumpkin": [{ "direction": 0 }, { "direction": 1 }, { "direction": 2 }, { "direction": 3 }],
    "minecraft:birch_pressure_plate": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:yellow_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:quartz_block": [{ "pillar_axis": "y", "chisel_type": "default" }, { "pillar_axis": "y", "chisel_type": "chiseled" }, { "pillar_axis": "y", "chisel_type": "lines" }, { "pillar_axis": "y", "chisel_type": "smooth" }, { "pillar_axis": "x", "chisel_type": "default" }, { "pillar_axis": "x", "chisel_type": "chiseled" }, { "pillar_axis": "x", "chisel_type": "lines" }, { "pillar_axis": "x", "chisel_type": "smooth" }, { "pillar_axis": "z", "chisel_type": "default" }, { "pillar_axis": "z", "chisel_type": "chiseled" }, { "pillar_axis": "z", "chisel_type": "lines" }, { "pillar_axis": "z", "chisel_type": "smooth" }, { "pillar_axis": "y", "chisel_type": "default" }, { "pillar_axis": "y", "chisel_type": "chiseled" }, { "pillar_axis": "y", "chisel_type": "lines" }, { "pillar_axis": "y", "chisel_type": "smooth" }],
    "minecraft:stripped_jungle_log": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:cobblestone_wall": [{ "wall_connection_type_west": "tall", "wall_connection_type_south": "tall", "wall_connection_type_east": "tall", "wall_connection_type_north": "tall", "wall_post_bit": true, "wall_block_type": "cobblestone" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "mossy_cobblestone" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "granite" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "diorite" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "andesite" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "sandstone" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "brick" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "stone_brick" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "mossy_stone_brick" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "nether_brick" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "end_brick" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "prismarine" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "red_sandstone" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "red_nether_brick" }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false, "wall_block_type": "cobblestone" }],
    "minecraft:respawn_anchor": [{ "respawn_anchor_charge": 0 }, { "respawn_anchor_charge": 1 }, { "respawn_anchor_charge": 2 }, { "respawn_anchor_charge": 3 }, { "respawn_anchor_charge": 4 }],
    "minecraft:sponge": [{ "sponge_type": "dry" }, { "sponge_type": "wet" }],
    "minecraft:red_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:concrete": [{ "color": "white" }, { "color": "orange" }, { "color": "magenta" }, { "color": "light_blue" }, { "color": "yellow" }, { "color": "lime" }, { "color": "pink" }, { "color": "gray" }, { "color": "silver" }, { "color": "cyan" }, { "color": "purple" }, { "color": "blue" }, { "color": "brown" }, { "color": "green" }, { "color": "red" }, { "color": "black" }],
    "minecraft:green_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:stone": [{ "stone_type": "stone" }, { "stone_type": "granite" }, { "stone_type": "granite_smooth" }, { "stone_type": "diorite" }, { "stone_type": "diorite_smooth" }, { "stone_type": "andesite" }, { "stone_type": "andesite_smooth" }],
    "minecraft:dark_oak_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:smoker": [{ "facing_direction": "south" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }, { "facing_direction": "down" }],
    "minecraft:crimson_door": [{ "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }],
    "minecraft:stained_glass": [{ "color": "white" }, { "color": "orange" }, { "color": "magenta" }, { "color": "light_blue" }, { "color": "yellow" }, { "color": "lime" }, { "color": "pink" }, { "color": "gray" }, { "color": "silver" }, { "color": "cyan" }, { "color": "purple" }, { "color": "blue" }, { "color": "brown" }, { "color": "green" }, { "color": "red" }, { "color": "black" }],
    "minecraft:unlit_redstone_torch": [{ "torch_facing_direction": "unknown" }, { "torch_facing_direction": "west" }, { "torch_facing_direction": "east" }, { "torch_facing_direction": "north" }, { "torch_facing_direction": "south" }, { "torch_facing_direction": "top" }],
    "minecraft:frame": [{ "item_frame_photo_bit": false, "item_frame_map_bit": false, "facing_direction": "east" }, { "item_frame_photo_bit": false, "item_frame_map_bit": false, "facing_direction": "west" }, { "item_frame_photo_bit": false, "item_frame_map_bit": false, "facing_direction": "south" }, { "item_frame_photo_bit": false, "item_frame_map_bit": false, "facing_direction": "north" }, { "item_frame_photo_bit": false, "item_frame_map_bit": true, "facing_direction": "east" }, { "item_frame_photo_bit": false, "item_frame_map_bit": true, "facing_direction": "west" }, { "item_frame_photo_bit": false, "item_frame_map_bit": true, "facing_direction": "south" }, { "item_frame_photo_bit": false, "item_frame_map_bit": true, "facing_direction": "north" }, { "item_frame_photo_bit": false, "item_frame_map_bit": false, "facing_direction": "down" }],
    "minecraft:carpet": [{ "color": "white" }, { "color": "orange" }, { "color": "magenta" }, { "color": "light_blue" }, { "color": "yellow" }, { "color": "lime" }, { "color": "pink" }, { "color": "gray" }, { "color": "silver" }, { "color": "cyan" }, { "color": "purple" }, { "color": "blue" }, { "color": "brown" }, { "color": "green" }, { "color": "red" }, { "color": "black" }],
    "minecraft:deepslate_brick_wall": [{ "wall_connection_type_west": "tall", "wall_connection_type_south": "tall", "wall_connection_type_east": "tall", "wall_connection_type_north": "tall", "wall_post_bit": true }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false }],
    "minecraft:normal_stone_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:lava_cauldron": [{ "cauldron_liquid": "water", "fill_level": 0 }, { "cauldron_liquid": "water", "fill_level": 1 }, { "cauldron_liquid": "water", "fill_level": 2 }, { "cauldron_liquid": "water", "fill_level": 3 }, { "cauldron_liquid": "water", "fill_level": 4 }, { "cauldron_liquid": "water", "fill_level": 5 }, { "cauldron_liquid": "water", "fill_level": 6 }],
    "minecraft:andesite_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:unpowered_repeater": [{ "repeater_delay": 0, "direction": 0 }, { "repeater_delay": 0, "direction": 1 }, { "repeater_delay": 0, "direction": 2 }, { "repeater_delay": 0, "direction": 3 }, { "repeater_delay": 1, "direction": 0 }, { "repeater_delay": 1, "direction": 1 }, { "repeater_delay": 1, "direction": 2 }, { "repeater_delay": 1, "direction": 3 }, { "repeater_delay": 2, "direction": 0 }, { "repeater_delay": 2, "direction": 1 }, { "repeater_delay": 2, "direction": 2 }, { "repeater_delay": 2, "direction": 3 }, { "repeater_delay": 3, "direction": 0 }, { "repeater_delay": 3, "direction": 1 }, { "repeater_delay": 3, "direction": 2 }, { "repeater_delay": 3, "direction": 3 }],
    "minecraft:stone_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:vine": [{ "vine_direction_bits": 0 }, { "vine_direction_bits": 1 }, { "vine_direction_bits": 2 }, { "vine_direction_bits": 3 }, { "vine_direction_bits": 4 }, { "vine_direction_bits": 5 }, { "vine_direction_bits": 6 }, { "vine_direction_bits": 7 }, { "vine_direction_bits": 8 }, { "vine_direction_bits": 9 }, { "vine_direction_bits": 10 }, { "vine_direction_bits": 11 }, { "vine_direction_bits": 12 }, { "vine_direction_bits": 13 }, { "vine_direction_bits": 14 }, { "vine_direction_bits": 15 }],
    "minecraft:waxed_oxidized_cut_copper_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:cyan_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:soul_torch": [{ "torch_facing_direction": "west" }, { "torch_facing_direction": "unknown" }],
    "minecraft:hay_block": [{ "pillar_axis": "y", "deprecated": 0 }, { "pillar_axis": "y", "deprecated": 1 }, { "pillar_axis": "y", "deprecated": 2 }, { "pillar_axis": "y", "deprecated": 3 }, { "pillar_axis": "x", "deprecated": 0 }, { "pillar_axis": "x", "deprecated": 1 }, { "pillar_axis": "x", "deprecated": 2 }, { "pillar_axis": "x", "deprecated": 3 }, { "pillar_axis": "z", "deprecated": 0 }, { "pillar_axis": "z", "deprecated": 1 }, { "pillar_axis": "z", "deprecated": 2 }, { "pillar_axis": "z", "deprecated": 3 }, { "pillar_axis": "y", "deprecated": 0 }, { "pillar_axis": "y", "deprecated": 1 }, { "pillar_axis": "y", "deprecated": 2 }, { "pillar_axis": "y", "deprecated": 3 }],
    "minecraft:darkoak_wall_sign": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:cauldron": [{ "cauldron_liquid": "water", "fill_level": 0 }, { "cauldron_liquid": "water", "fill_level": 1 }, { "cauldron_liquid": "water", "fill_level": 2 }, { "cauldron_liquid": "water", "fill_level": 3 }, { "cauldron_liquid": "water", "fill_level": 4 }, { "cauldron_liquid": "water", "fill_level": 5 }, { "cauldron_liquid": "water", "fill_level": 6 }],
    "minecraft:acacia_wall_sign": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:kelp": [{ "liquid_depth": 0 }, { "kelp_age": 1 }, { "liquid_depth": 0 }, { "kelp_age": 3 }, { "kelp_age": 4 }, { "kelp_age": 5 }, { "kelp_age": 6 }, { "kelp_age": 7 }, { "kelp_age": 8 }, { "kelp_age": 9 }, { "kelp_age": 10 }, { "kelp_age": 11 }, { "kelp_age": 12 }, { "kelp_age": 13 }, { "kelp_age": 14 }, { "kelp_age": 15 }, { "kelp_age": 0 }],
    "minecraft:mangrove_log": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:acacia_standing_sign": [{ "ground_sign_direction": 0 }, { "ground_sign_direction": 1 }, { "ground_sign_direction": 2 }, { "ground_sign_direction": 3 }, { "ground_sign_direction": 4 }, { "ground_sign_direction": 5 }, { "ground_sign_direction": 6 }, { "ground_sign_direction": 7 }, { "ground_sign_direction": 8 }, { "ground_sign_direction": 9 }, { "ground_sign_direction": 10 }, { "ground_sign_direction": 11 }, { "ground_sign_direction": 12 }, { "ground_sign_direction": 13 }, { "ground_sign_direction": 14 }, { "ground_sign_direction": 15 }],
    "minecraft:warped_wall_sign": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:bone_block": [{ "pillar_axis": "y", "deprecated": 0 }, { "pillar_axis": "y", "deprecated": 1 }, { "pillar_axis": "y", "deprecated": 2 }, { "pillar_axis": "y", "deprecated": 3 }, { "pillar_axis": "x", "deprecated": 0 }, { "pillar_axis": "x", "deprecated": 1 }, { "pillar_axis": "x", "deprecated": 2 }, { "pillar_axis": "x", "deprecated": 3 }, { "pillar_axis": "z", "deprecated": 0 }, { "pillar_axis": "z", "deprecated": 1 }, { "pillar_axis": "z", "deprecated": 2 }, { "pillar_axis": "z", "deprecated": 3 }, { "pillar_axis": "y", "deprecated": 0 }, { "pillar_axis": "y", "deprecated": 1 }, { "pillar_axis": "y", "deprecated": 2 }, { "pillar_axis": "y", "deprecated": 3 }],
    "minecraft:light_blue_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:stripped_oak_log": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:cyan_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:green_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:orange_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:activator_rail": [{ "rail_direction": 0, "rail_data_bit": false }, { "rail_direction": 1, "rail_data_bit": false }, { "rail_direction": 2, "rail_data_bit": false }, { "rail_direction": 3, "rail_data_bit": false }, { "rail_direction": 4, "rail_data_bit": false }, { "rail_direction": 5, "rail_data_bit": false }],
    "minecraft:stone_block_slab4": [{ "stone_slab_type_4": "mossy_stone_brick", "top_slot_bit": false }, { "stone_slab_type_4": "smooth_quartz", "top_slot_bit": false }, { "stone_slab_type_4": "stone", "top_slot_bit": false }, { "stone_slab_type_4": "cut_sandstone", "top_slot_bit": false }, { "stone_slab_type_4": "cut_red_sandstone", "top_slot_bit": false }],
    "minecraft:stripped_crimson_hyphae": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:waxed_exposed_cut_copper_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:potatoes": [{ "growth": 0 }, { "growth": 1 }, { "growth": 2 }, { "growth": 3 }, { "growth": 4 }, { "growth": 5 }, { "growth": 6 }, { "growth": 7 }],
    "minecraft:heavy_weighted_pressure_plate": [{ "redstone_signal": 0 }, { "redstone_signal": 1 }, { "redstone_signal": 2 }, { "redstone_signal": 3 }, { "redstone_signal": 4 }, { "redstone_signal": 5 }, { "redstone_signal": 6 }, { "redstone_signal": 7 }, { "redstone_signal": 8 }, { "redstone_signal": 9 }, { "redstone_signal": 10 }, { "redstone_signal": 11 }, { "redstone_signal": 12 }, { "redstone_signal": 13 }, { "redstone_signal": 14 }, { "redstone_signal": 15 }],
    "minecraft:crimson_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:darkoak_standing_sign": [{ "ground_sign_direction": 0 }, { "ground_sign_direction": 1 }, { "ground_sign_direction": 2 }, { "ground_sign_direction": 3 }, { "ground_sign_direction": 4 }, { "ground_sign_direction": 5 }, { "ground_sign_direction": 6 }, { "ground_sign_direction": 7 }, { "ground_sign_direction": 8 }, { "ground_sign_direction": 9 }, { "ground_sign_direction": 10 }, { "ground_sign_direction": 11 }, { "ground_sign_direction": 12 }, { "ground_sign_direction": 13 }, { "ground_sign_direction": 14 }, { "ground_sign_direction": 15 }],
    "minecraft:cake": [{ "bite_counter": 0 }, { "bite_counter": 1 }, { "bite_counter": 2 }, { "bite_counter": 3 }, { "bite_counter": 4 }, { "bite_counter": 5 }, { "bite_counter": 6 }],
    "minecraft:oak_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:barrel": [{ "open_bit": false, "facing_direction": "down" }, { "open_bit": false, "facing_direction": "up" }, { "open_bit": false, "facing_direction": "north" }, { "open_bit": false, "facing_direction": "south" }, { "open_bit": false, "facing_direction": "west" }, { "open_bit": false, "facing_direction": "east" }],
    "minecraft:coral_fan": [{ "liquid_depth": 0 }, { "coral_fan_direction": 0, "coral_color": "pink" }, { "liquid_depth": 0 }, { "coral_fan_direction": 0, "coral_color": "red" }, { "coral_fan_direction": 0, "coral_color": "yellow" }],
    "minecraft:stripped_dark_oak_log": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:powered_comparator": [{ "output_lit_bit": false, "output_subtract_bit": false, "direction": 0 }, { "output_lit_bit": false, "output_subtract_bit": false, "direction": 1 }, { "output_lit_bit": false, "output_subtract_bit": false, "direction": 2 }, { "output_lit_bit": false, "output_subtract_bit": false, "direction": 3 }, { "output_lit_bit": false, "output_subtract_bit": true, "direction": 0 }, { "output_lit_bit": false, "output_subtract_bit": true, "direction": 1 }, { "output_lit_bit": false, "output_subtract_bit": true, "direction": 2 }, { "output_lit_bit": false, "output_subtract_bit": true, "direction": 3 }, { "output_lit_bit": true, "output_subtract_bit": false, "direction": 0 }, { "output_lit_bit": true, "output_subtract_bit": false, "direction": 1 }, { "output_lit_bit": true, "output_subtract_bit": false, "direction": 2 }, { "output_lit_bit": true, "output_subtract_bit": false, "direction": 3 }, { "output_lit_bit": true, "output_subtract_bit": true, "direction": 0 }, { "output_lit_bit": true, "output_subtract_bit": true, "direction": 1 }, { "output_lit_bit": true, "output_subtract_bit": true, "direction": 2 }, { "output_lit_bit": true, "output_subtract_bit": true, "direction": 3 }],
    "minecraft:muddy_mangrove_roots": [{ "pillar_axis": "y" }, { "pillar_axis": "x" }, { "pillar_axis": "z" }],
    "minecraft:lime_glazed_terracotta": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:scaffolding": [{ "stability_check": false, "stability": 0 }, { "stability_check": false, "stability": 1 }, { "stability_check": false, "stability": 2 }, { "stability_check": false, "stability": 3 }, { "stability_check": false, "stability": 4 }, { "stability_check": false, "stability": 5 }, { "stability_check": false, "stability": 6 }, { "stability_check": false, "stability": 7 }, { "stability_check": true, "stability": 0 }, { "stability_check": true, "stability": 1 }, { "stability_check": true, "stability": 2 }, { "stability_check": true, "stability": 3 }, { "stability_check": true, "stability": 4 }, { "stability_check": true, "stability": 5 }, { "stability_check": true, "stability": 6 }, { "stability_check": true, "stability": 7 }],
    "minecraft:lava": [{ "liquid_depth": 0 }, { "liquid_depth": 1 }, { "liquid_depth": 2 }, { "liquid_depth": 3 }, { "liquid_depth": 4 }, { "liquid_depth": 5 }, { "liquid_depth": 6 }, { "liquid_depth": 7 }, { "liquid_depth": 8 }, { "liquid_depth": 9 }, { "liquid_depth": 10 }, { "liquid_depth": 11 }, { "liquid_depth": 12 }, { "liquid_depth": 13 }, { "liquid_depth": 14 }, { "liquid_depth": 15 }],
    "minecraft:polished_blackstone_brick_wall": [{ "wall_connection_type_west": "tall", "wall_connection_type_south": "tall", "wall_connection_type_east": "tall", "wall_connection_type_north": "tall", "wall_post_bit": true }, { "wall_connection_type_west": "none", "wall_connection_type_south": "none", "wall_connection_type_east": "none", "wall_connection_type_north": "none", "wall_post_bit": false }],
    "minecraft:acacia_fence_gate": [{ "open_bit": false, "in_wall_bit": false, "direction": 0 }, { "open_bit": false, "in_wall_bit": false, "direction": 1 }, { "open_bit": false, "in_wall_bit": false, "direction": 2 }, { "open_bit": false, "in_wall_bit": false, "direction": 3 }, { "open_bit": true, "in_wall_bit": false, "direction": 0 }, { "open_bit": true, "in_wall_bit": false, "direction": 1 }, { "open_bit": true, "in_wall_bit": false, "direction": 2 }, { "open_bit": true, "in_wall_bit": false, "direction": 3 }, { "open_bit": false, "in_wall_bit": true, "direction": 0 }, { "open_bit": false, "in_wall_bit": true, "direction": 1 }, { "open_bit": false, "in_wall_bit": true, "direction": 2 }, { "open_bit": false, "in_wall_bit": true, "direction": 3 }, { "open_bit": true, "in_wall_bit": true, "direction": 0 }, { "open_bit": true, "in_wall_bit": true, "direction": 1 }, { "open_bit": true, "in_wall_bit": true, "direction": 2 }, { "open_bit": true, "in_wall_bit": true, "direction": 3 }],
    "minecraft:command_block": [{ "facing_direction": "down", "conditional_bit": false }, { "facing_direction": "up", "conditional_bit": false }, { "facing_direction": "north", "conditional_bit": false }, { "facing_direction": "south", "conditional_bit": false }, { "facing_direction": "west", "conditional_bit": false }, { "facing_direction": "east", "conditional_bit": false }],
    "minecraft:black_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:gray_candle_cake": [{ "lit": false }, { "lit": true }],
    "minecraft:prismarine_bricks_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:loom": [{ "direction": 0 }, { "direction": 1 }, { "direction": 2 }, { "direction": 3 }],
    "minecraft:carved_pumpkin": [{ "direction": 0 }, { "direction": 1 }, { "direction": 2 }, { "direction": 3 }],
    "minecraft:double_stone_block_slab3": [{ "stone_slab_type_3": "end_stone_brick", "top_slot_bit": false }, { "stone_slab_type_3": "smooth_red_sandstone", "top_slot_bit": false }, { "stone_slab_type_3": "polished_andesite", "top_slot_bit": false }, { "stone_slab_type_3": "andesite", "top_slot_bit": false }, { "stone_slab_type_3": "diorite", "top_slot_bit": false }, { "stone_slab_type_3": "polished_diorite", "top_slot_bit": false }, { "stone_slab_type_3": "granite", "top_slot_bit": false }, { "stone_slab_type_3": "polished_granite", "top_slot_bit": false }, { "stone_slab_type_3": "end_stone_brick", "top_slot_bit": true }, { "stone_slab_type_3": "smooth_red_sandstone", "top_slot_bit": true }, { "stone_slab_type_3": "polished_andesite", "top_slot_bit": true }, { "stone_slab_type_3": "andesite", "top_slot_bit": true }, { "stone_slab_type_3": "diorite", "top_slot_bit": true }, { "stone_slab_type_3": "polished_diorite", "top_slot_bit": true }, { "stone_slab_type_3": "granite", "top_slot_bit": true }, { "stone_slab_type_3": "polished_granite", "top_slot_bit": true }],
    "minecraft:standing_sign": [{ "ground_sign_direction": 0 }, { "ground_sign_direction": 1 }, { "ground_sign_direction": 2 }, { "ground_sign_direction": 3 }, { "ground_sign_direction": 4 }, { "ground_sign_direction": 5 }, { "ground_sign_direction": 6 }, { "ground_sign_direction": 7 }, { "ground_sign_direction": 8 }, { "ground_sign_direction": 9 }, { "ground_sign_direction": 10 }, { "ground_sign_direction": 11 }, { "ground_sign_direction": 12 }, { "ground_sign_direction": 13 }, { "ground_sign_direction": 14 }, { "ground_sign_direction": 15 }],
    "minecraft:sapling": [{ "sapling_type": "oak", "age_bit": false }, { "sapling_type": "spruce", "age_bit": false }, { "sapling_type": "birch", "age_bit": false }, { "sapling_type": "jungle", "age_bit": false }, { "sapling_type": "acacia", "age_bit": false }, { "sapling_type": "dark_oak", "age_bit": false }],
    "minecraft:polished_deepslate_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:crimson_wall_sign": [{ "facing_direction": "down" }, { "facing_direction": "up" }, { "facing_direction": "north" }, { "facing_direction": "south" }, { "facing_direction": "west" }, { "facing_direction": "east" }],
    "minecraft:birch_trapdoor": [{ "upside_down_bit": false, "open_bit": false, "direction": 0 }, { "upside_down_bit": false, "open_bit": false, "direction": 1 }, { "upside_down_bit": false, "open_bit": false, "direction": 2 }, { "upside_down_bit": false, "open_bit": false, "direction": 3 }, { "upside_down_bit": true, "open_bit": false, "direction": 0 }, { "upside_down_bit": true, "open_bit": false, "direction": 1 }, { "upside_down_bit": true, "open_bit": false, "direction": 2 }, { "upside_down_bit": true, "open_bit": false, "direction": 3 }, { "upside_down_bit": false, "open_bit": true, "direction": 0 }, { "upside_down_bit": false, "open_bit": true, "direction": 1 }, { "upside_down_bit": false, "open_bit": true, "direction": 2 }, { "upside_down_bit": false, "open_bit": true, "direction": 3 }, { "upside_down_bit": true, "open_bit": true, "direction": 0 }, { "upside_down_bit": true, "open_bit": true, "direction": 1 }, { "upside_down_bit": true, "open_bit": true, "direction": 2 }, { "upside_down_bit": true, "open_bit": true, "direction": 3 }],
    "minecraft:black_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:stone_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:polished_blackstone_brick_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:coral": [{ "coral_color": "blue", "dead_bit": false }, { "coral_color": "pink", "dead_bit": false }, { "coral_color": "purple", "dead_bit": false }, { "coral_color": "red", "dead_bit": false }, { "coral_color": "yellow", "dead_bit": false }],
    "minecraft:leaves2": [{ "new_leaf_type": "acacia", "update_bit": false, "persistent_bit": true }, { "new_leaf_type": "dark_oak", "update_bit": false, "persistent_bit": true }],
    "minecraft:sweet_berry_bush": [{ "growth": 0 }, { "growth": 1 }, { "growth": 2 }, { "growth": 3 }, { "growth": 4 }, { "growth": 5 }, { "growth": 6 }, { "growth": 7 }],
    "minecraft:trapdoor": [{ "upside_down_bit": false, "open_bit": false, "direction": 0 }, { "upside_down_bit": false, "open_bit": false, "direction": 1 }, { "upside_down_bit": false, "open_bit": false, "direction": 2 }, { "upside_down_bit": false, "open_bit": false, "direction": 3 }, { "upside_down_bit": true, "open_bit": false, "direction": 0 }, { "upside_down_bit": true, "open_bit": false, "direction": 1 }, { "upside_down_bit": true, "open_bit": false, "direction": 2 }, { "upside_down_bit": true, "open_bit": false, "direction": 3 }, { "upside_down_bit": false, "open_bit": true, "direction": 0 }, { "upside_down_bit": false, "open_bit": true, "direction": 1 }, { "upside_down_bit": false, "open_bit": true, "direction": 2 }, { "upside_down_bit": false, "open_bit": true, "direction": 3 }, { "upside_down_bit": true, "open_bit": true, "direction": 0 }, { "upside_down_bit": true, "open_bit": true, "direction": 1 }, { "upside_down_bit": true, "open_bit": true, "direction": 2 }, { "upside_down_bit": true, "open_bit": true, "direction": 3 }],
    "minecraft:spruce_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:gray_candle": [{ "candles": 0, "lit": false }, { "candles": 1, "lit": false }, { "candles": 2, "lit": false }, { "candles": 3, "lit": false }, { "candles": 0, "lit": true }, { "candles": 1, "lit": true }, { "candles": 2, "lit": true }, { "candles": 3, "lit": true }],
    "minecraft:jungle_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:deepslate_tile_slab": [{ "top_slot_bit": false }, { "top_slot_bit": true }],
    "minecraft:cactus": [{ "age": 0 }, { "age": 1 }, { "age": 2 }, { "age": 3 }, { "age": 4 }, { "age": 5 }, { "age": 6 }, { "age": 7 }, { "age": 8 }, { "age": 9 }, { "age": 10 }, { "age": 11 }, { "age": 12 }, { "age": 13 }, { "age": 14 }, { "age": 15 }],
    "minecraft:coral_fan_dead": [{ "coral_fan_direction": 0, "coral_color": "blue" }, { "coral_fan_direction": 0, "coral_color": "pink" }, { "coral_fan_direction": 0, "coral_color": "purple" }, { "coral_fan_direction": 0, "coral_color": "red" }, { "coral_fan_direction": 0, "coral_color": "yellow" }],
    "minecraft:dark_oak_button": [{ "facing_direction": "down", "button_pressed_bit": false }, { "facing_direction": "up", "button_pressed_bit": false }, { "facing_direction": "north", "button_pressed_bit": false }, { "facing_direction": "south", "button_pressed_bit": false }, { "facing_direction": "west", "button_pressed_bit": false }, { "facing_direction": "east", "button_pressed_bit": false }],
    "minecraft:jungle_door": [{ "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": false, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": false, "door_hinge_bit": true }, { "open_bit": false, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": false, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": false, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 0, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 1, "upper_block_bit": true, "door_hinge_bit": true }, { "open_bit": true, "direction": 2, "upper_block_bit": true, "door_hinge_bit": false }, { "open_bit": true, "direction": 3, "upper_block_bit": true, "door_hinge_bit": true }],
    "minecraft:nether_brick_stairs": [{ "weirdo_direction": 0, "upside_down_bit": false }, { "weirdo_direction": 1, "upside_down_bit": false }, { "weirdo_direction": 2, "upside_down_bit": false }, { "weirdo_direction": 3, "upside_down_bit": false }, { "weirdo_direction": 0, "upside_down_bit": true }, { "weirdo_direction": 1, "upside_down_bit": true }, { "weirdo_direction": 2, "upside_down_bit": true }, { "weirdo_direction": 3, "upside_down_bit": true }],
    "minecraft:snow_layer": [{ "height": 0, "covered_bit": false }, { "height": 1, "covered_bit": false }, { "height": 2, "covered_bit": false }, { "height": 3, "covered_bit": false }, { "height": 4, "covered_bit": false }, { "height": 5, "covered_bit": false }, { "height": 6, "covered_bit": false }, { "height": 7, "covered_bit": false }, { "height": 0, "covered_bit": true }, { "height": 1, "covered_bit": true }, { "height": 2, "covered_bit": true }, { "height": 3, "covered_bit": true }, { "height": 4, "covered_bit": true }, { "height": 5, "covered_bit": true }, { "height": 6, "covered_bit": true }, { "height": 7, "covered_bit": true }]
};
